//_________________________________________________________________________
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// NOMBRE: GBADDraw.c
// AUTOR: Eduardo Pinera Aznarez
// DESCRIPCION: Funciones 2D para Game Boy Advanced.
//
// FECHA INICIO: 16 de Noviembre de 2000.
// ULTIMA ACTUALIZACION: 16 de Noviembre de 2000.
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//

#include <stdio.h>
#include <stdlib.h>

#include "GBASystem.h"
#include "GBADDraw.h"



//_________________________________________________________________________
//*************************************************************************
//*************************************************************************
//
//
//
//							CONVERT	XX BPP to 32 BPP.
//
//
//
//*************************************************************************
//*************************************************************************
//



//
//			      VARIABLES
//____________________________________________




//
//		LINEAS DE PROTOTIPO DE FUNCIONES
//____________________________________________
DWORD Convert8to32(BYTE R, BYTE G, BYTE B);
DWORD Convert15to32(WORD Color);
DWORD Convert16to32(WORD Color);
DWORD Convert24to32(DWORD Color);



//
//			      FUNCIONES
//____________________________________________


//________________________________________________________________________
// Nombre: Convert8to32() 
// Descripcion: Convierte los componentes RGB de 8Bpp a 32Bpp
//
DWORD Convert8to32(BYTE R, BYTE G, BYTE B)
{
	DWORD Result = 0x00000000;
		
	Result = ((R << 16) | (G << 8) | B);

	return Result;
}



//________________________________________________________________________
// Nombre: Convert15to32()
// Descripcion: Convierte los componentes RGB de 15Bpp a 32Bpp
//
DWORD Convert15to32(WORD Color)
{
	DWORD Result = 0x00000000;
	DWORD R = 0x00000000;
	DWORD G = 0x00000000;
	DWORD B = 0x00000000;

	R = ((Color >> 10) & 31);
	R = ((R << 3) | (R >> 2));

	G = ((Color >> 5) & 31);
	G = ((G << 3) | (G >> 2));

	B = (Color & 31);
	B = ((B << 3) | (B >> 2));

	Result = ((R << 16) | (G << 8) | B);

	return Result;
}


//________________________________________________________________________
// Nombre: Convert16to32()
// Descripcion: Convierte los componentes RGB de 16Bpp a 32Bpp.
//
DWORD Convert16to32(WORD Color)
{
	DWORD Result = 0x00000000;
	DWORD R = 0x00000000;
	DWORD G = 0x00000000;
	DWORD B = 0x00000000;

	R = ((Color >> 11) & 31);
	R = ((R << 3) | (R >> 2));

	G = ((Color >> 6) & 63);
	G = ((G << 2) | (G >> 4));

	B = (Color & 31);
	B = ((B << 3) | (B >> 2));

	Result = ((R << 16) | (G << 8) | B);

	return Result;
}


//________________________________________________________________________
// Nombre: Convert24to32()
// Descripcion: Convierte los componentes RGB de 24Bpp a 32Bpp.
//
DWORD Convert24to32(DWORD Color)
{
	DWORD Result = 0x00000000;	

	Result = (Color & 0x00FFFFFF);

	return Result;
}






//_________________________________________________________________________
//*************************************************************************
//*************************************************************************
//
//
//
//							CONVERT 32 BPP to XX BPP.
//
//
//
//*************************************************************************
//*************************************************************************
//


//
//			      VARIABLES
//____________________________________________




//
//		LINEAS DE PROTOTIPO DE FUNCIONES
//____________________________________________
VOID Convert32to8(DWORD Color, BYTE *R, BYTE *G, BYTE *B);
WORD Convert32to15(DWORD Color);
WORD Convert32to16(DWORD Color);
DWORD Convert32to24(DWORD Color);


//
//				FUNCIONES
//____________________________________________

//________________________________________________________________________
// Nombre: Convert32to8() 
// Descripcion: Convierte los componentes RGB de 8Bpp a 32Bpp
//
VOID Convert32to8(DWORD Color, BYTE *R, BYTE *G, BYTE *B)
{
	*R = 0x00;
	*G = 0x00;
	*B = 0x00;

	*R = (BYTE) ((Color >> 16) & 255);
	*G = (BYTE) ((Color >> 8) & 255);
	*B = (BYTE) (Color & 255);
}

//________________________________________________________________________
// Nombre: Convert32to15()
// Descripcion: Convierte los componentes RGB de 32Bpp a 15Bpp.
//
WORD Convert32to15(DWORD Color)
{
	DWORD Result = 0x00000000;
	DWORD R = 0x00000000;
	DWORD G = 0x00000000;
	DWORD B = 0x00000000;

	R = ((Color >> 16) & 255);
	G = ((Color >> 8) & 255);
	B = (Color & 255);

	R = R >> 3;
	G = G >> 3;
	B = B >> 3;

	Result = ((R << 10) | (G << 5) | B);

	return (WORD)Result;
}


//________________________________________________________________________
// Nombre: Convert32to16()
// Descripcion: Convierte los componentes RGB de 32Bpp a 16Bpp.
//
WORD Convert32to16(DWORD Color)
{
	DWORD Result = 0x00000000;
	DWORD R = 0x00000000;
	DWORD G = 0x00000000;
	DWORD B = 0x00000000;

	R = ((Color >> 16) & 255);
	G = ((Color >> 8) & 255);
	B = (Color & 255);

	R = R >> 3;
	G = G >> 2;
	B = B >> 3;

	Result = ((R << 11) | (G << 5) | B);

	return (WORD)Result;
}


//________________________________________________________________________
// Nombre: Convert32to24()
// Descripcion: Convierte los componentes RGB de 32Bpp a 24Bpp.
//
DWORD Convert32to24(DWORD Color)
{
	DWORD Result = 0x00000000;

	Result = (Color & 0x00FFFFFF);

	return Result;
}




//_________________________________________________________________________
//*************************************************************************
//*************************************************************************
//
//
//
//					VARIABLES Y FUNCIONES GENERICAS DE DIBUJO Y 
//							VOLCADO DE PIXELS Y BUFFERS
//								   PARA EL MODO 3
//
//
//
//*************************************************************************
//*************************************************************************
//


//
//			      VARIABLES
//____________________________________________




//
//		LINEAS DE PROTOTIPO DE FUNCIONES
//____________________________________________
void M3PutPixelRGB888(unsigned char X, unsigned char Y, unsigned char ColorR, unsigned char ColorG, unsigned char ColorB);
void M3PutPixel888(unsigned char X, unsigned char Y, unsigned long DWColor);
void M3PutPixelRGB555(unsigned char X, unsigned char Y, unsigned char ColorR, unsigned char ColorG, unsigned char ColorB);
void M3PutPixel555(unsigned char X, unsigned char Y, unsigned short Color);

void M3Blit15(long X, long Y, unsigned char Width, unsigned char Height, unsigned short *Buffer);

void M3ClearScreen(void);


//
//				FUNCIONES
//____________________________________________


//________________________________________________________________________
// Name: M3PutPixel888()
// Desc: Dibuja un Pixel suministrandole los componentes con 
//	una profundidad de color 8:8:8 que sera convertida a 5:5:5.
//
void M3PutPixelRGB888(unsigned char X, unsigned char Y, unsigned char ColorR, unsigned char ColorG, unsigned char ColorB)
{
	unsigned short Color;

	Color  = ((ColorB << 5) >> 8);
	ColorB = Color;

	Color  = ((ColorG << 5) >> 8);
	ColorG = Color;

	Color  = ((ColorR << 5) >> 8);
	ColorR = Color;
	
	Color  = ColorB;
	Color  = (Color << 5);
	Color |= ColorG;
	Color  = (Color << 5);
	Color |= ColorR;

	//FASTEST VERSION by epaedu@netcom.es
	__asm
	{
			mov		r2, #0

			mov		r1, Y
			mov		r1, r1, lsl #8

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #7

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #6

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #5

			add		r2, r2, r1			

			mov		r1, X

			add		r2, r2, r1
			add		r2, r2, r1

			mov		r0, #0x06000000
			add		r0, r0, r2

			mov		r1, Color
			strh	r1, [r0], 1
	}
}


//________________________________________________________________________
// Name: M3PutPixel888()
// Desc: Dibuja un Pixel suministrandole los componentes en una 
//	doble palabra con una profundidad de color 8:8:8 que sera 
//	convertida a 5:5:5.
//
void M3PutPixel888(unsigned char X, unsigned char Y, unsigned long DWColor)
{
	unsigned short Color;

	Color = Convert32to15(DWColor);

	//FASTEST VERSION by epaedu@netcom.es
	__asm
	{
			mov		r2, #0

			mov		r1, Y
			mov		r1, r1, lsl #8

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #7

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #6

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #5

			add		r2, r2, r1			

			mov		r1, X

			add		r2, r2, r1
			add		r2, r2, r1

			mov		r0, #0x06000000
			add		r0, r0, r2

			mov		r1, Color
			strh	r1, [r0], 1
	}
}


//________________________________________________________________________
// Name: M3PutPixelRGB555()
// Desc: Dibuja un Pixel suministrandole los componentes con 
//	una profundidad de color 5:5:5.
//
void M3PutPixelRGB555(unsigned char X, unsigned char Y, unsigned char ColorR, unsigned char ColorG, unsigned char ColorB)
{
	unsigned short Color;

	if(ColorB > 63)
	{
		Color  = ((ColorB << 5) >> 8);
		ColorB = Color;
	}

	if(ColorG > 63)
	{
		Color  = ((ColorG << 5) >> 8);
		ColorG = Color;
	}

	if(ColorR > 63)
	{
		Color  = ((ColorR << 5) >> 8);
		ColorR = Color;
	}
	
	Color  = ColorB;
	Color  = (Color << 5);
	Color |= ColorG;
	Color  = (Color << 5);
	Color |= ColorR;

	//FASTEST VERSION by epaedu@netcom.es
	__asm
	{
			mov		r2, #0

			mov		r1, Y
			mov		r1, r1, lsl #8

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #7

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #6

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #5

			add		r2, r2, r1			

			mov		r1, X

			add		r2, r2, r1
			add		r2, r2, r1

			mov		r0, #0x06000000
			add		r0, r0, r2

			mov		r1, Color
			strh	r1, [r0], 1
	}
}



//________________________________________________________________________
// Name: M3PutPixel555()
// Desc: Dibuja un Pixel suministrandole los componentes en una palabra
//	con una profundidad de color 5:5:5.
//
void M3PutPixel555(unsigned char X, unsigned char Y, unsigned short Color)
{
	//FASTEST VERSION by epaedu@netcom.es
	__asm
	{
			mov		r2, #0

			mov		r1, Y
			mov		r1, r1, lsl #8

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #7

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #6

			add		r2, r2, r1

			mov		r1, Y
			mov		r1, r1, lsl #5

			add		r2, r2, r1

			mov		r1, X

			add		r2, r2, r1
			add		r2, r2, r1

			mov		r0, #0x06000000
			add		r0, r0, r2

			mov		r1, Color
			strh	r1, [r0], 1
	}
}



//________________________________________________________________________
// Name: M3Blit15()
// Desc: Vuelca un buffer a la pantalla de la Game Boy Advanced.
//
void M3Blit15(long X, long Y, unsigned char Width, unsigned char Height, unsigned short *Buffer)
{
	GBARECT BufferRect;
	GBARECT ScreenRect;

	unsigned char	CounterX;
	unsigned char	CounterY;
	unsigned long	BufferOffset;	

	// if the entery buffer is out of screen, return;
	if((X < 0) && ((X+Width) < 0)){ return; }
	if(X > 239){ return; }

	if((Y < 0) && ((Y+Height) < 0)){ return; }
	if(Y > 159){ return; }


	if((X > -1) && ((X+Width) < 240) && (Y > -1) && ((Y+Height) < 160))
	{
		__asm
		{
			// r0 = SCREEN + (y*240*2)+(x*2)
			mov		r2, #0
			mov		r1, Y
			mov		r1, r1, lsl #8
			add		r2, r2, r1
			mov		r1, Y
			mov		r1, r1, lsl #7
			add		r2, r2, r1
			mov		r1, Y
			mov		r1, r1, lsl #6
			add		r2, r2, r1
			mov		r1, Y
			mov		r1, r1, lsl #5
			add		r2, r2, r1
			mov		r1, X
			add		r2, r2, r1
			add		r2, r2, r1
			mov		r0, #0x06000000
			add		r0, r0, r2

			// r3 = (240*2) - (x*2)
			mov		r2, Buffer
			mov		r3, #480
			mov		r4, Width
			sub		r3, r3, r4
			sub		r3, r3, r4

			mov		r4, Height
			FullYloop:

				mov		r5, Width
				FullXloop:
					// Move a half-word (16 bits) from [r2] to [r0]
					ldrh	r1, [r2]
					strh	r1, [r0]

					add		r2, r2, #2
					add		r0, r0, #2

					// The 's'-suffix is neccecary to update the flags, or you'd get an
					// eternal loop..
					subs	r5, r5, #1
				bne FullXloop

				add		r0, r0, r3
				subs	r4, r4, #1
			bne FullYloop
		}
	}
	else
	{
		BufferRect.Left		= X;
		BufferRect.Top		= Y;
		BufferRect.Right	= (X + Width);
		BufferRect.Bottom	= (Y + Height);

		ScreenRect.Left		= X;
		ScreenRect.Top		= Y;
		ScreenRect.Right	= (X + Width);
		ScreenRect.Bottom	= (Y + Height);

		if(X < 0)
		{
			ScreenRect.Left		= 0;
			BufferRect.Left		= (X * -1);

			if((X+Width) > 239)
			{
				ScreenRect.Right	= 240;
				BufferRect.Right	= (BufferRect.Left + 240);
			}
			else
			{
				ScreenRect.Right	= (X + Width);
				BufferRect.Right	= BufferRect.Left + (X + Width);
			}
		}
		else
		{
			ScreenRect.Left		= X;
			BufferRect.Left		= 0;

			if((X+Width) > 239)
			{
				ScreenRect.Right	= 240;
				BufferRect.Right	= (240 - X);
			}
			else
			{
				ScreenRect.Right	= (X + Width);
				BufferRect.Right	= Width;
			}
		}


		if(Y < 0)
		{
			ScreenRect.Top		= 0;
			BufferRect.Top		= (Y * -1);

			if((Y+Height) > 159)
			{
				ScreenRect.Bottom	= 160;
				BufferRect.Bottom	= (BufferRect.Top + 160);
			}
			else
			{
				ScreenRect.Bottom	= (Y + Height);
				BufferRect.Bottom	= BufferRect.Top + (Y + Height);
			}
		}
		else
		{
			ScreenRect.Top		= Y;
			BufferRect.Top		= 0;

			if((Y+Height) > 159)
			{
				ScreenRect.Bottom	= 160;
				BufferRect.Bottom	= (160 - Y);
			}
			else
			{
				ScreenRect.Bottom	= (Y + Height);
				BufferRect.Bottom	= Height;
			}
		}

		for(CounterY = (BYTE) ScreenRect.Top; CounterY < (BYTE)ScreenRect.Bottom; CounterY++)
		{
			BufferOffset = ((BufferRect.Top << 7)+(BufferRect.Top << 6)+(BufferRect.Top << 5)+(BufferRect.Top << 4))+(BufferRect.Left);
			BufferRect.Top++;

			for(CounterX = (BYTE) ScreenRect.Left; CounterX < (BYTE)ScreenRect.Right; CounterX++)
			{
				M3PutPixel555(CounterX, CounterY, Buffer[BufferOffset++]);
			}
		}
	}
}


//________________________________________________________________________
// Name: M3ClearScreen()
// Desc: Limpia la pantalla de la Game Boy Advanced.
//
void M3ClearScreen(void) 
{
	__asm
	{
		mov		r0, #0x06000000
		mov		r1, #0x00000000
		mov		r2, #19200
		FullCleanLoop:

			str		r1, [r0]
			add		r0, r0, #4

			subs	r2, r2, #1
		bne FullCleanLoop
	}
}
